"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.main import Queue
from qme.main.executor import get_named_executor
from qme.logger import bot
import sys
import os


def main(args, extra):

    result = None
    if not args.actions:
        sys.exit("Please provide an executor to list actions for, or an action.")

    # User has provided an executor to list actions for
    elif len(args.actions) == 1:

        # Assume looking for executor first, fall back to action on last job
        try:
            executor = get_named_executor(args.actions[0])
            actions = [[x] for x in executor.get_actions()]
            bot.table(actions)
        except:
            queue = Queue(config_dir=args.config_dir)
            task = queue.get()
            result = task.run_action(args.actions[0])

    # NEED TO DEBUG THIS FOR SLURM
    # user provided exec <action> <taskid>
    elif len(args.actions) == 2:
        taskid, action = args.actions
        queue = Queue(config_dir=args.config_dir)
        task = queue.get(taskid)
        result = task.run_action(action)

    # Print result to terminal
    if result:
        if isinstance(result, list):
            result = " ".join(result)
        print(result)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.main import Queue
from qme.logger import bot
import sys
import os


def main(args, extra):

    # Create a queue object, run the command to match to an executor
    queue = Queue(config_dir=args.config_dir)

    # Case 1: empty list indicates listing all
    if not args.executor:
        bot.table(queue.list())
    else:
        # Each in the list can be a full executor or a task id
        for executor in args.executor:
            bot.table(queue.list(executor))



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.utils.file import (
    write_json,
    mkdir_p,
    read_json,
    recursive_find,
    get_latest_modified,
)
from qme.main.database.base import Database
from qme.main.executor import get_named_executor
from qme.logger import bot
from glob import glob
import shutil
import uuid
import os
import re
import sys


class FileSystemDatabase(Database):
    """A FileSystemDatabase writes raw json to files at $HOME/.qme/database
       This is the default flat database for qme, and on init we ensure
       that the database folder is created in QME_HOME.
    """

    database = "filesystem"

    def __init__(self, config_dir, config=None, **kwargs):
        """init for the filesystem ensures that the base folder (named 
           according to the studyid) exists.
        """
        self.config = config
        self.create_database(config_dir)

    def create_database(self, config_dir):
        """Create the database. The parent folder must exist.
        """
        self.data_base = os.path.abspath(os.path.join(config_dir, "database"))
        if not os.path.exists(config_dir):
            sys.exit(f"{config_dir} must exist to create database there.")
        if not os.path.exists(self.data_base):
            mkdir_p(self.data_base)

    # Global

    def clear(self):
        """clear (delete) all tasks.
        """
        for executor_dir in self.iter_executors(fullpath=True):
            if os.path.exists(executor_dir):
                bot.info(f"Removing {executor_dir}")
                shutil.rmtree(executor_dir)
        return True

    # Add or Update requires executor

    def add_task(self, executor):
        """Create a filesystem task based on an executor type. The executor controls
           what data is exported and the uid, the task object just handles saving it.
        """
        return FileSystemTask(executor, data_base=self.data_base)

    def update_task(self, executor, updates=None):
        """update a task with a json dictionary.
        """
        task = FileSystemTask(executor, exists=True, data_base=self.data_base)
        task.update({"data": executor.export(), "command": executor.command})

    # Get, delete, etc. only require taskid

    def get_task(self, taskid=None):
        """Get a task based on a taskid. Exits on error if doesn't exist. If
           a task id is not provided, get the last run task.
        """
        if not taskid:
            taskid = os.path.basename(
                get_latest_modified(self.data_base, pattern="*.json")
            ).replace(".json", "")
        executor = taskid.split("-", 1)[0]
        executor = get_named_executor(executor, taskid, config=self.config)
        return FileSystemTask(executor, exists=True, data_base=self.data_base)

    def delete_task(self, taskid):
        """delete a task based on a specific task id. All task ids must be
           in the format of <taskid>-<uid> without extra dashes so we can
           reliably split based on the first dash.
        """
        task = self.get_task(taskid)
        if not task:
            bot.error(f"{taskid} does not exist in the database.")
            return False
        os.remove(task.filename)
        bot.info(f"{taskid} has been removed.")
        return True

    def delete_executor(self, name):
        """delete all tasks for an executor, based on executor's name (str).
        """
        executor_dir = os.path.join(self.data_base, name)
        if not os.path.exists(executor_dir):
            bot.error(f"Executor {executor} does not exist.")
            return False
        shutil.rmtree(executor_dir)
        return True

    def iter_executors(self, fullpath=False):
        """list executors based on the subfolders in the base database folder.
        """
        for contender in os.listdir(self.data_base):
            contender = os.path.join(self.data_base, contender)
            if os.path.isdir(contender):
                if not fullpath:
                    yield os.path.basename(contender)
                else:
                    yield contender

    def list_tasks(self, name=None):
        """list tasks, either under a particular executor name (if provided)
           or just the executors. This returns tasks in rows to be printed
           (or otherwise parsed).
        """
        listpath = self.data_base
        if name:
            listpath = os.path.join(listpath, name)
        rows = []
        for filename in recursive_find(listpath, pattern="*.json"):
            rows.append([os.path.basename(filename).replace(".json", "")])
        return rows


class FileSystemTask:
    """A Filesystem Task can take a task id, determine if the task exists,
       and then interact with the data. If the task is instantiated without
       a taskid it is assumed to not exist yet, otherwise it must already
       exist.
    """

    def __init__(self, executor, data_base, exists=False):
        """A FileSystem task tasks some task id and command for an executor.
           We provide a simple interface to retrieve the data file, and 
           do an initial creation if it doesn't exist.

           Arguments:
             taskid (str) : the executor-uuid for the task
             command (list) : the command to be executed
             data_base (str) : the path where the database exists.
             exists (bool) : if True, must already exists (default is False)
        """
        self.taskid = executor.taskid
        self.executor = executor
        self.data_base = data_base
        self.data = {}
        self.create(exists)

    @property
    def filename(self):
        return "%s.json" % os.path.join(self.executor_dir, self.executor.taskid)

    @property
    def executor_dir(self):
        return os.path.join(self.data_base, self.executor.name)

    def update(self, updates=None):
        """Update a data file. This means reading, updating, and writing.
        """
        updates = updates or {}
        if updates:
            self.data.update(updates)
            self.save()

    def create(self, should_exist=False):
        """create the filename if it doesn't exist, otherwise if it should (and
           does not) exit on error.
        """
        if should_exist:
            if not os.path.exists(self.filename):

                # Might be provided prefix
                contenders = glob(
                    "%s*" % os.path.join(self.executor_dir, self.executor.taskid)
                )
                if len(contenders) == 1:
                    self.executor.taskid = re.sub(
                        "(%s/|[.]json)"
                        % os.path.join(self.data_base, self.executor.name),
                        "",
                        contenders[0],
                    )

                elif len(contenders) > 1:
                    sys.exit(f"More than one task found for {self.executor.taskid}")
                else:
                    sys.exit(f"Cannot find task {self.executor.taskid}")
            self.data = self.load()

        if not os.path.exists(self.executor_dir):
            os.mkdir(self.executor_dir)

        # If it's the first time saving, create basic file
        if not should_exist:
            self.data = {
                "executor": self.executor.name,
                "uid": self.executor.taskid,
                "command": self.executor.command,
                "data": self.executor.export(),
            }
            self.save()

    def export(self):
        """wrapper to expose the executor.export function
        """
        return self.executor.export()

    def save(self):
        """Save a json object to the task.
        """
        write_json(self.data, self.filename)

    def summary(self):
        return self.executor.summary()

    def load(self):
        """Given a task, load data from filename.
        """
        if os.path.exists(self.filename):
            return read_json(self.filename)

    def run_action(self, name, **kwargs):
        """Run an action, meaning running the executor's run_action but
           providing data from the database.
        """
        return self.executor.run_action(name, self.data, **kwargs)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from .shell import ShellExecutor
from .slurm import SlurmExecutor
import sys
import re


def matches(Executor, command):
    """Given a command, determine if it matches the regular expression
       that determines to use the executor or not. This applies to all
       executors except for the shell executor. This means that all non-shell
       classes need to have a matchstring defined.
    """
    if not hasattr(Executor, "matchstring"):
        raise NotImplementedError

    if isinstance(command, list):
        command = " ".join(command)
    return not re.search(Executor.matchstring, command) == None


def get_executor(command=None, config=None):
    """get executor will return the correct executor depending on a command (or
       other string) matching a regular expression. If nothing matches, we 
       default to a shell executor. Each non-shell executor should expose
       a common "matches" function (provided by the base class) that will
       handle parsing the command (a list) to a single string, and checking
       if it matches a regular expression.
    """
    # Slurm Executor
    if matches(SlurmExecutor, command):
        executor = SlurmExecutor(command=command)

    # Default is standard shell command
    else:
        executor = ShellExecutor(command=command)
    executor.config = config
    return executor


def get_named_executor(name, taskid=None, config=None):
    """get a named executor, meaning determining based on name and not command
    """
    if name == "shell":
        executor = ShellExecutor(taskid)
    elif name == "slurm":
        executor = SlurmExecutor(taskid)
    else:
        sys.exit(f"{name} is not a known executor.")
    executor.config = config
    return executor



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

__version__ = "0.0.11"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsochat@stanford.edu"
NAME = "qme"
PACKAGE_URL = "http://www.github.com/vsoch/qme"
KEYWORDS = "python,dashboard,jobs,actions"
DESCRIPTION = "job submission and dashboard library"
LICENSE = "LICENSE"

################################################################################
# Global requirements


INSTALL_REQUIRES = ()

APP_REQUIRES = (
    ("Flask", {"min_version": "1.0.2"}),
    ("Flask-SocketIO", {"min_version": "3.1.2"}),
    ("flask-restful", {"min_version": "0.3.8"}),
    ("gevent-websocket", {"min_version": "0.10.1"}),
)

DATABASE_REQUIRES = (("sqlalchemy", {"min_version": None}),)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)


ALL_REQUIRES = INSTALL_REQUIRES + DATABASE_REQUIRES + APP_REQUIRES



#!/usr/bin/env python
"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import sys
import pytest


def test_executors_filesystem(tmp_path):
    """test each executor type with the filesystem
    """
    from qme.main import Queue

    config_dir = os.path.join(str(tmp_path), ".qme")
    queue = Queue(config_dir=config_dir)

    # The following commands should map to the following executors
    commands = ["ls"]
    executors = ["shell"]
    for i, command in enumerate(commands):

        # The executor name should be chosen based on the command
        name = executors[i]
        task = queue.run("ls")
        assert task.executor.name == name
        assert task.filename == os.path.join(
            queue.config_dir, "database", name, "%s.json" % task.taskid
        )
        assert task.summary()

        # Task.load includes the file dump, the upper level keys should be same
        content = task.load()
        for key in ["executor", "uid", "data", "command"]:
            assert key in content

        # Task.export includes the executor specific data
        data = task.export()
        for key in ["pwd", "user", "timestamp"]:
            assert key in data

        # Ensure a message is added, if defined with run
        task = queue.run("ls", message="this is my custom message")
        assert "message" in task.load()["data"]


def test_filesystem(tmp_path):
    """Test loading and using a queue with the filesystem database.
    """
    from qme.main import Queue

    config_dir = os.path.join(str(tmp_path), ".qme")
    queue = Queue(config_dir=config_dir)

    assert os.path.exists(config_dir)
    assert queue.config_dir == config_dir
    assert queue.config.configfile == os.path.join(queue.config_dir, "config.ini")
    assert queue.database == "filesystem"
    assert queue.db.database == "filesystem"
    assert queue.db.data_base == os.path.join(config_dir, "database")

    with pytest.raises(SystemExit):
        queue.search("Not possible for non-relational databases.")

    # Test list, empty without anything
    assert not queue.list()

    # Run a task
    task = queue.run("ls")
    assert task.executor.name == "shell"
    assert task.taskid.startswith("shell")
    assert len(queue.list()) == 1

    # Rerun the task, should still only have one
    rerun = queue.rerun()
    assert rerun.taskid == task.taskid
    assert len(queue.list()) == 1

    # queue.get should return last task, given no id
    lasttask = queue.get()
    assert lasttask.taskid == task.taskid

    # Run a new task
    newtask = queue.run("whoami")
    assert len(queue.list()) == 2
    exports = newtask.export()

    # Check exports
    for required in ["pwd", "output", "error", "cmd", "returncode"]:
        assert required in exports
    assert exports["pwd"] == os.getcwd()
    assert exports["cmd"] == ["whoami"]
    assert exports["returncode"] == 0

    # Get a task id that isn't the last task
    notlast = queue.get(task.taskid)
    assert task.taskid == notlast.taskid

    # Clean up a specific task (no prompt)
    queue.clear(task.taskid, noprompt=True)
    assert len(queue.list()) == 1
    queue.clear(noprompt=True)
    assert not queue.list()



#!/usr/bin/env python
"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import sys
import pytest


def test_executor_shell(tmp_path):
    """Test the shell executor. A shell executor should have added a returncode,
       output, and error to it's data export.
    """
    from qme.main import Queue

    config_dir = os.path.join(str(tmp_path), ".qme")
    queue = Queue(config_dir=config_dir)
    task = queue.run("ls")
    assert task.executor.name == "shell"
    assert task.filename == os.path.join(
        queue.config_dir, "database", "shell", "%s.json" % task.taskid
    )
    assert task.summary()

    # Task.load includes the file dump, the upper level keys should be same
    content = task.load()
    for key in ["executor", "uid", "data"]:
        assert key in content

    # Task.export includes the executor specific data
    data = task.export()
    for key in ["cmd", "pwd", "output", "error", "returncode"]:
        assert key in data



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from flask import Flask, render_template
from flask_socketio import emit

# from werkzeug import secure_filename
from qme.app.server import app, socketio
from qme.defaults import QME_SOCKET_UPDATE_SECONDS

from time import sleep
from threading import Thread, Event

# thread for updates
thread = Thread()
thread_stop_event = Event()


## Main Index View


@app.route("/")
def index():

    # Return view based on database type
    if app.queue.database == "filesystem":
        return render_template(
            "home/filesystem-index.html", database=app.queue.database
        )
    return render_template("home/relational-index.html", database=app.queue.database)


## Updating database rows


def update_database():
    """A function to be run at some interval to update the qme database listing
    """
    while not thread_stop_event.isSet():

        # The data sent to the table depends on the database
        if app.queue.database == "filesystem":

            # Break into executor types and taskids
            message = "use a relational or sqlite database to see more metadata."
            rows = [(x[0].split("-")[0], x[0], message) for x in app.queue.list()]
            socketio.emit(
                "FSdatabase",
                {"rows": rows, "database": app.queue.database},
                namespace="/update",
            )

        # sqlite or other relational
        else:
            rows = [(x[0].split("-")[0], x[0], x[1]) for x in app.queue.list()]
            socketio.emit(
                "RELdatabase",
                {"rows": rows, "database": app.queue.database},
                namespace="/update",
            )

        socketio.sleep(QME_SOCKET_UPDATE_SECONDS)


@socketio.on("deleterow", namespace="/table/action")
def delete_row(json):
    """a request to delete a particular row"""
    app.logger.debug("Received deletion request for %s" % json.get("taskid"))
    taskid = json.get("taskid", "doesnotexist")
    was_deleted = app.queue.clear(target=taskid, noprompt=True)
    socketio.emit(
        "deleterowcomplete",
        {"wasdeleted": was_deleted, "taskid": taskid},
        namespace="/table/action",
    )


@socketio.on("rerunrow", namespace="/table/action")
def rerun_row(json):
    """a request to re-run a particular task.
    """
    app.logger.debug("Received re-run request for %s" % json.get("taskid"))
    taskid = json.get("taskid", "doesnotexist")
    was_rerun = app.queue.rerun(taskid) is not None
    socketio.emit(
        "reruncomplete",
        {"wasrerun": was_rerun, "taskid": taskid},
        namespace="/table/action",
    )


@socketio.on("connect", namespace="/update")
def update_connect():
    # need visibility of the global thread object
    global thread
    app.logger.debug("Client connected")

    # Start the process to update the table
    app.logger.debug("Starting Thread")
    thread = socketio.start_background_task(update_database)


@socketio.on("disconnect", namespace="/update")
def update_disconnect():
    app.logger.debug("Client disconnected")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.main.database.relational import RelationalDatabase
from qme.logger import bot
import os


class SqliteDatabase(RelationalDatabase):
    """A SqliteDatabase writes to a qme.db file in $HOME/.qme. This is
       the suggested database backend for QueueMe, as it doesn't require anything
       beyond a filesystem and still allows for relational type queries.
    """

    database = "sqlite"

    def __init__(self, config_dir, config=None, **kwargs):
        """init for the filesystem ensures that the base folder (named 
           according to the studyid) exists.
        """
        database_file = kwargs.get("database_string", "qme.db") or "qme.db"

        # Derive database path, use default of qme.db if not provided
        db_path = os.path.join(config_dir, database_file)
        if not db_path.endswith(".db"):
            bot.exit(
                f"Invalid database file for sqlite, {database_file} does not end in .db"
            )

        self.config = config
        self.db = "sqlite:///%s" % db_path
        self.create_database()



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.main import Queue
import shlex
import sys
import os


def run(args, extra):

    # Create a queue object, run the command to match to an executor
    queue = Queue(config_dir=args.config_dir)

    # Don't allow an empty command!
    if not args.cmd:
        sys.exit("Please provide a command for QueueMe to execute.")

    command = args.cmd

    # --help needs to be quoted, make sure if provided, gets parsed into command
    if any(["--help" in x for x in args.cmd]):
        command = []
        for item in args.cmd:
            command += shlex.split(item)

    # Extra might include unparsed arguments
    queue.run(command=command + extra, message=args.message)


def rerun(args, extra):

    # Create a queue object, run the command to match to an executor
    queue = Queue(config_dir=args.config_dir)
    queue.rerun(taskid=args.taskid, message=args.message)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.utils.file import read_file, get_user

from datetime import datetime
import os
import re
import tempfile
import subprocess
import uuid


class Capturing:
    """capture output from stdout and stderr into capture object.
       This is based off of github.com/vsoch/gridtest but modified
       to write files. The stderr and stdout are set to temporary files at
       the init of the capture, and then they are closed when we exit. This
       means expected usage looks like:

       with Capturing() as capture:
           process = subprocess.Popen(...)
           
       And then the output and error are retrieved from reading the files:
       and exposed as properties to the client:

           capture.out
           capture.err

       And cleanup means deleting these files, if they exist.
    """

    def __enter__(self):
        self.set_stdout()
        self.set_stderr()
        self.output = []
        self.error = []
        return self

    def set_stdout(self):
        self.stdout = open(tempfile.mkstemp()[1], "w")

    def set_stderr(self):
        self.stderr = open(tempfile.mkstemp()[1], "w")

    def __exit__(self, *args):
        self.stderr.close()
        self.stdout.close()

    @property
    def out(self):
        """Return output stream. Returns empty string if empty or doesn't exist.
           Returns (str) : output stream written to file
        """
        if os.path.exists(self.stdout.name):
            return read_file(self.stdout.name)
        return ""

    @property
    def err(self):
        """Return error stream. Returns empty string if empty or doesn't exist.
           Returns (str) : error stream written to file
        """
        if os.path.exists(self.stderr.name):
            return read_file(self.stderr.name)
        return ""

    def cleanup(self):
        for filename in [self.stdout.name, self.stderr.name]:
            if os.path.exists(filename):
                os.remove(filename)


class ExecutorBase:
    """A qme executor exists to translate a terminal command into a parsed
       job (shown in the dashboard) and expose one or more actions for it.
       The base executor will work for any generic command, generates
       status based on return codes, and exposes basic options to cancel (kill)
       or re-run.
    """

    name = "base"

    def __init__(self, taskid):
        """set a unique id that includes executor name (type) and random uuid)
        """
        uid = str(uuid.uuid4())
        if taskid:
            _, uid = taskid.split("-", 1)
        self.taskid = "%s-%s" % (self.name, uid)
        self.pwd = os.getcwd()
        self.message = None
        self.actions = {}
        if not hasattr(self, "data"):
            self.data = {}

    def _export_common(self):
        """export common task variables such as present working directory, user,
           and timestamp for when it was run. This might include envars at some
           point, but we'd need to be careful.
        """
        common = {
            "pwd": self.pwd,
            "user": get_user(),
            "timestamp": str(datetime.now()),
        }
        if self.message:
            common["message"] = self.message
        return common

    def export(self):
        """return data as json. This is intended to save to the task database.
           Any important executor specific metadata should be added to self.data
        """
        # Get common context (e.g., pwd)
        common = self._export_common()
        common.update(self.data)
        return common

    @property
    def command(self):
        raise NotImplementedError

    def run_action(self, name, data, **kwargs):
        """Check for a named action in the executors list.
           This is called from the queue that can also add the data for the task
           as "data." The user should be able to run an action by name, e.g.,
           executor.action('status', data) and take key word arguments, which
           is exposed by a task as task.run_action('status', data)
        """
        if name in self.actions:
            return self.actions[name](data, **kwargs)

    def get_actions(self):
        """return list of actions to expose"""
        return list(self.actions)

    def capture(self, cmd):
        """capture is a helper function to capture a shell command. We
           use Capturing and then save attributes like the pid, output, error
           to it, and return to the calling function. For example:

           capture = self.capture_command(cmd)
           self.pid = capture.pid
           self.returncode = capture.returncode
           self.out = capture.output
           self.err = capture.error
        """
        # Capturing provides temporary output and error files
        with Capturing() as capture:
            process = subprocess.Popen(
                cmd,
                stdout=capture.stdout,
                stderr=capture.stderr,
                universal_newlines=True,
            )
            capture.pid = process.pid
            returncode = process.poll()

            # Iterate through the output
            while returncode is None:
                returncode = process.poll()

        # Get the remainder of lines, add return code
        capture.output += [x for x in self.decode(capture.out) if x]
        capture.error += [x for x in self.decode(capture.err) if x]

        # Cleanup capture files and save final return code
        capture.cleanup()
        capture.returncode = returncode
        return capture

    def get_setting(self, key, default=None):
        """Get a setting, meaning that we first check the environment, then
           the config file, and then (if provided) a default.
        """
        # First preference to environment
        envar = ("QME_%s_%s" % (self.name, key)).upper()
        envar = os.environ.get(envar)
        if envar is not None:
            return envar

        # Next preference to config setting
        executor = "executor.%s" % self.name
        if executor not in self.config.config:
            return default
        if key in self.config.config[executor]:
            return self.config.get(executor, key)
        return default

    def summary(self):
        return "[%s]" % self.name

    def execute(self, cmd=None, message=None):
        raise NotImplementedError

    def get_output(self):
        raise NotImplementedError

    def get_error(self):
        raise NotImplementedError



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import locale
import os
import shlex
import shutil
import sys

from .base import ExecutorBase


class ShellExecutor(ExecutorBase):
    """A shell executor is the most basic of executors to run some shell command.
       We use the default functions provided by the BaseExecutor class to store
       the command, run, get the return code, and retry.
    """

    name = "shell"

    def __init__(self, taskid=None, command=None):
        super().__init__(taskid)
        self.reset(command)

    def summary(self):
        if self.data.get("returncode") is not None:
            return "[%s][returncode: %s]" % (self.taskid, self.data.get("returncode"))
        return "[%s][%s]" % (self.name, self.command)

    def reset(self, command=None):
        """refresh output and error streams
        """
        self.data = {
            "output": [],
            "error": [],
            "returncode": None,
            "pid": None,
            "cmd": [],
            "status": None,
        }
        if command:
            self.set_command(command)

    @property
    def command(self):
        return " ".join(self.data.get("cmd", []))

    def set_command(self, cmd):
        """parse is called when a new command is provided to ensure we have
           a list. We don't check that the executable is on the path,
           as the initialization might not occur in the runtime environment.
        """
        if not isinstance(cmd, list):
            cmd = shlex.split(cmd)
        self.data["cmd"] = cmd

    def execute(self, cmd=None, message=None):
        """Execute a system command and return output and error. Execute
           should take a cmd (a string or list) and execute it according to
           the executor. Attributes should be set on the class that are
           added to self.export. Since the functions here are likely needed
           by most executors, we create a self._execute() class that is called
           instead, and can be used by the other executors.
        """
        self.message = message
        return self._execute(cmd)

    def _execute(self, cmd=None):
        """The actual class to do the execution - can be used if ShellExecutor
           is used as a super and the class using it defines a custom execute
        """
        # Reset the output and error records
        self.reset(cmd or self.data.get("cmd"))
        cmd = self.data.get("cmd")

        # The executable must be found, return code 1 if not
        executable = shutil.which(cmd[0])
        if not executable:
            self.data["error"] = ["%s not found." % cmd[0]]
            self.data["returncode"] = 1
            return self.data["output"], self.data["error"]

        # remove the original executable
        args = cmd[1:]
        self.data["status"] = "running"

        # Use updated command with executable and remainder (list)
        cmd = [executable] + args

        # Capturing provides temporary output and error files
        capture = self.capture(cmd)
        self.data["pid"] = capture.pid
        self.data["returncode"] = capture.returncode
        self.data["output"] = capture.output
        self.data["error"] = capture.error
        self.data["status"] = "complete"
        return self.data["output"], self.data["error"]

    def decode(self, line):
        """Given a line of output (error or regular) decode using the
           system default, if appropriate
        """
        loc = locale.getdefaultlocale()[1]

        try:
            line = line.decode(loc)
        except:
            pass
        return line

    def get_output(self):
        """Returns the output from shell command
        :rtype: str
        """
        return self.data.get("output")

    def get_error(self):
        """Returns the error from shell command
        :rtype: str
        """
        return self.data.get("error")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from flask_socketio import SocketIO
from flask import Flask

from qme.logger import bot
from qme.defaults import QME_HOSTNAME

import logging
import os


class QueueMeServer(Flask):
    def __init__(self, *args, **kwargs):
        super(QueueMeServer, self).__init__(*args, **kwargs)


app = QueueMeServer(__name__)
app.config.from_object("qme.app.config")

# turn the flask app into a socketio app
socketio = SocketIO(app, async_mode=None, logger=True, engineio_logger=True)

from qme.app.views import *
from qme.app.api import *


def start(port=5000, debug=True, queue=None, host=None):
    """Start can be invoked when this file is executed (see __main__ below)
       or used as a function to programmatically start a server. If started
       via qme view, we can add the queue to the server. If you want to change
       the hostname, set the environment variable QME_HOSTNAME or set on command
       line with qme start.
    """
    host = host or QME_HOSTNAME
    bot.info(f"QueueMe: running on http://{host}:{port}")

    # If the user doesn't specify a queue, use default
    if not queue:
        from qme.main import Queue

        queue = Queue()

    # Customize the logger to log to the app folder
    file_handler = logging.FileHandler(os.path.join(queue.config_dir, "dashboard.log"))
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)
    app.logger.addHandler(file_handler)
    app.logger.setLevel(logging.DEBUG)

    # Add the queue to the server
    app.queue = queue
    socketio.run(app, port=port, debug=debug, host=host)


# This is how the command line version will run
if __name__ == "__main__":
    start()



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.main import Queue
import sys
import os


def main(args, extra):

    # Clear an executor, taskid, or target
    queue = Queue(config_dir=args.config_dir)
    queue.clear(args.target, noprompt=args.force)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import sys
import re

from qme.utils.file import read_file
from qme.logger import bot
from .shell import ShellExecutor


class SlurmExecutor(ShellExecutor):
    """A slurm executor parses an srun command and exposes options for getting
       a job status.
    """

    name = "slurm"
    matchstring = "^sbatch"

    def __init__(self, taskid=None, command=None):
        super().__init__(taskid, command)
        self.actions = {
            "status": self.action_get_status,
            "output": self.action_get_output,
            "error": self.action_get_error,
            "cancel": self.action_cancel,
        }

    def execute(self, cmd=None, message=None):
        """Execute a system command and return output and error. Execute
           should take a cmd (a string or list) and execute it according to
           the executor. Attributes should be set on the class that are
           added to self.export. Since the functions here are likely needed
           by most executors, we create a self._execute() class that is called
           instead, and can be used by the other executors.
        """
        self.message = message
        self._execute(cmd)
        if self.data["returncode"] == 0:

            # Find the job id, and any --out or --error files
            match = re.search("[0-9]+", self.data["output"][0])
            if not match:
                bot.exit(f"Unable to derive job id from {self.data['output']}")
            self.data["jobid"] = match.group()

            # Get output file, or default to $PWD/slurm-<jobid>
            self.data["errorfile"] = os.path.join(
                self.pwd, "slurm-%s.err" % self.data["jobid"]
            )
            self.data["outputfile"] = os.path.join(
                self.pwd, "slurm-%s.out" % self.data["jobid"]
            )

            match = re.search(r"(--out|-o) (?P<output>[^\s-]+)", self.command)
            if match:
                self.data["outputfile"] = match.groups("output")[1]

            match = re.search(r"(--err|-e) (?P<error>[^\s-]+)", self.command)
            if match:
                self.data["errorfile"] = match.groups("error")[1]

    # Actions

    def action_get_status(self, data):
        """Get the status with squeue, given a jobid. This returns information
           for the job that is based on a format string, default looks like:

		{'jobid': '904366',
		 'jobname': 'run_job.sh',
		 'partition': 'owners',
		 'alloccpus': '1',
		 'elapsed': '00:00:00',
		 'state': 'PENDING',
		 'exitcode': '0:0'}
        """
        fmt = self.get_setting(
            "sacct_format", "jobid,jobname,partition,alloccpus,elapsed,state,exitcode"
        )
        jobid = data.get("jobid")
        if jobid:
            capture = self.capture(
                ["sacct", "--job", jobid, "--format=%s" % fmt, "--noheader"]
            )
            values = {}
            output = [x for x in capture.output[0].strip().split(" ") if x]
            for i, varname in enumerate(fmt.split(",")):
                values[varname] = output[i]
            return values

    def action_get_output(self, data):
        """Get error stream, if the file exists.
        """
        outputfile = data.get("outputfile")
        if os.path.exists(outputfile):
            return read_file(outputfile)
        return ["%s does not exist.\n" % outputfile]

    def action_get_error(self, data):
        """Get just output stream, if the file exists.
        """
        errorfile = data.get("outputfile")
        if os.path.exists(errorfile):
            return read_file(errorfile)
        return ["%s does not exist.\n" % errorfile]

    def action_cancel(self, data):
        """Cancel a job if there is a jobid
        """
        jobid = data.get("jobid")
        if jobid:
            capture = self.capture(["scancel", jobid])
            return capture.output



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

uuid_regex = "[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}"



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.logger import bot
import json
import sys


try:
    from sqlalchemy import Column, DateTime, Integer, String, Text, ForeignKey, func
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.orm import relationship, backref
    from uuid import uuid4
except:
    sys.exit("sqlalchemy is required for a non-filesystem database.")

# Shared declarative base
Base = declarative_base()


class Task(Base):
    """An executor task. The id is prefixed with the executor type, and must
       be unique.
    """

    __tablename__ = "task"
    taskid = Column(String(150), primary_key=True)
    command = Column(String(500))
    timestamp = Column(DateTime, default=func.now())
    executor_name = Column(String(50))
    data = Column(Text, nullable=True)

    def __init__(self, taskid=None, executor=None, command=None):
        self.taskid = taskid
        self.executor_name = executor
        self.command = command

    def summary(self):
        return self.executor.summary()

    def load(self):
        """loading a task means exporting as json"""
        data = {
            "executor": self.executor_name,
            "uid": self.taskid,
            "data": {},
            "command": self.command,
        }

        if self.data:
            data["data"] = json.loads(self.data)

        return data

    def export(self):
        """Export removes the outer wrapper, and just returns the data"""
        return self.load().get("data", {})

    def run_action(self, name, **kwargs):
        """Run an action, meaning that we prepare data to it, and then run
           the self.executor.run_action(name, data) function.
        """
        return self.executor.run_action(name, self.export(), **kwargs)

    def __repr__(self):
        return "<Task %r>" % self.taskid



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.logger import bot
from qme.utils.file import mkdir_p, read_json, write_json

import configparser
import os
import shutil

here = os.path.dirname(os.path.abspath(__file__))


class Config:
    def __init__(self, config_dir=None, load=True):
        """Make sure the config file exists, along with the QueueMe Home. 
           If it doesn't exist, create it. Return the full path to the config
           file to store with the Queue and load the config.
        """
        from qme.defaults import QME_HOME

        QME_HOME = config_dir or QME_HOME
        if not os.path.exists(QME_HOME):
            bot.info(f"Creating QueueMe home at {QME_HOME}.")
            mkdir_p(QME_HOME)

        # Configfile is based in QME_HOME
        self.configfile = os.path.join(QME_HOME, "config.ini")

        # If the config file doesn't exist, generate it
        if not os.path.exists(self.configfile):
            default_config = os.path.join(here, "config.ini")
            shutil.copyfile(default_config, self.configfile)

        # Load the config file if wanted
        if load and self.configfile:
            self.read()

    def get(self, section, key):
        """A wrapper to config.get to directly interact with the self.config
        """
        return self.config.get(section, key)

    def update(self, section, key, value, save=False):
        """update client secrets will update the data structure for a particular
           authentication. This should only be used for a (quasi permanent) token
           or similar. The secrets file, if found, is updated and saved by default.
        """
        if section not in self.config:
            self.config[section] = {}
        self.config[section][key] = value

        # If save is true (to persist settings)
        if save:
            self.save()

    def save(self):
        """save configuration back to it's original file
        """
        with open(self.configfile, "w") as configfile:
            self.config.write(configfile)

    def read(self, configfile=None):
        """read in configuration file. By default use self.configfile.
        """
        configfile = configfile or self.configfile
        config = configparser.ConfigParser()
        config.read(configfile)
        self.config = config



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.main.database.base import Database
from qme.main.executor import get_named_executor
from qme.logger import bot

from sqlalchemy import create_engine, desc
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy import and_, or_, not_

import os
import json
import sys


class RelationalDatabase(Database):
    """A RelationalDatabase is a more robust relational datbase (to sqlite).
       Since the global database property can be any of postgresql, mysql+pysq;,
       it is defined on init. The sqlite database also uses this class, but defines
       a custom init function to handle the $HOME/.qme/qme.db file.
    """

    def __init__(self, config_dir, config=None, **kwargs):
        """init for the filesystem ensures that the base folder (named 
           according to the studyid) exists.
        """
        self.database = kwargs.get("database")
        self.config = config
        database_string = kwargs.get("database_string")
        if not database_string:
            sys.exit(
                "A database url must be defined to use a relational database. Set with qme config --database"
            )

        # The database url includes the type and string
        self.db = "%s://%s" % (self.database, database_string)
        self.create_database()

    def create_database(self):
        """create the databsae based on the string, whether it's relational or
           sqlite. self.db must be defined.
        """
        from qme.main.database.models import Base

        self.engine = create_engine(self.db)
        self.session = scoped_session(
            sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
        )
        Base.query = self.session.query_property()
        Base.metadata.create_all(bind=self.engine)
        self.Base = Base

    # Global

    def clear(self):
        """clear (delete) all tasks. This could be improved to cascade instead.
        """
        from qme.main.database.models import Task

        Task.query.delete()
        self.session.commit()
        return True

    # Add or Update requires executor

    def add_task(self, executor):
        """Create a new task based on an executor type. The executor controls
           what data is exported and the uid, the task object just handles saving it.
        """
        from qme.main.database.models import Task

        task = Task(taskid=executor.taskid, executor=executor.name)
        self.session.add(task)
        self.session.commit()
        task.executor = executor
        return task

    def update_task(self, executor, updates=None):
        """update a task with a json dictionary.
        """
        from qme.main.database.models import Task

        task = Task.query.filter(Task.taskid == executor.taskid).first()
        updates = updates or {}
        updates.update(executor.export())

        # If we find a matching task
        if task:

            # If a command is defined
            if executor.command:
                task.command = executor.command

            # Load the previous data to update
            data = {}
            if task.data:
                data = json.loads(task.data)
            data.update(updates)

            task.data = json.dumps(data)
            self.session.add(task)
            self.session.commit()
            return task

    # Get, delete, etc. only require taskid

    def get_task(self, taskid=None):
        """Get a task based on a taskid. Exits on error if doesn't exist. If
           a task id is not provided, get the last run task.
        """
        from qme.main.database.models import Task

        # Retrieve either the last task, or the one with a specific taskid
        if not taskid:
            task = self.session.query(Task).order_by(desc("timestamp")).first()
            if not task:
                sys.exit(f"There are no tasks in the database.")
        else:
            task = Task.query.filter(Task.taskid == taskid).first()

            # If an exact match isn't there, look for partial match
            if not task:
                query = "%" + taskid + "%"
                query = self.session.query(Task).filter(Task.taskid.ilike(query))
                results = self.session.execute(query).fetchall()
                if len(results) == 1:
                    return self.get_task(results[0][0])
                elif len(results) > 1:
                    sys.exit(f"More than one task found for {taskid}")
                else:
                    sys.exit(f"Cannot find task {taskid}")

        # Add the executor to the task
        executor = task.taskid.split("-", 1)[0]
        task.executor = get_named_executor(executor, task.taskid, config=self.config)
        return task

    def delete_task(self, taskid):
        """delete a task based on a specific task id. All task ids must be
           in the format of <taskid>-<uid> without extra dashes so we can
           reliably split based on the first dash.
        """
        from qme.main.database.models import Task

        task = self.get_task(taskid)
        if not task:
            bot.error(f"{taskid} does not exist in the database.")
            return False
        Task.query.filter(Task.taskid == task.taskid).delete()
        self.session.commit()
        bot.info(f"{taskid} has been removed.")
        return True

    def delete_executor(self, name):
        """delete all tasks for an executor, based on executor's name (str).
        """
        from qme.main.database.models import Task

        deleted_items = False
        for task in Task.query.filter(Task.executor_name == name):
            deleted_items = True
            self.session.delete(task)
        self.session.commit()
        return deleted_items

    def iter_executors(self, fullpath=False):
        """list executors based on the subfolders in the base database folder.
        """
        from qme.main.database.models import Task

        for executor in self.session.query(Task.executor_name).distinct():
            yield executor[0]

    def list_tasks(self, name=None):
        """list tasks, either under a particular executor name (if provided)
           or just the executors. This returns tasks in rows to be printed
           (or otherwise parsed).
        """
        from qme.main.database.models import Task

        if name:
            tasks = Task.query.filter(Task.executor_name == name)
        else:
            tasks = Task.query.all()

        rows = []
        for task in tasks:
            rows.append([task.taskid, task.command or ""])
        return rows

    def search(self, query):
        """Search across the database for a particular query.
        """
        from qme.main.database.models import Task

        # Ensure that query can be part of a larger string
        query = "%" + query + "%"

        query = self.session.query(Task).filter(
            or_(Task.command.ilike(query), Task.data.ilike(query))
        )
        # list of tuples, (taskid, command, datetime, executor]
        results = self.session.execute(query).fetchall()
        return [[r[0], str(r[2]), r[1]] for r in results]



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import flask
from flask_restful import Resource, Api
from qme.logger import bot
from qme.app.server import app
import os


def list_tasks(executor=None):
    """A shared function to list one or more tasks, optionally for an executor,
       and return a json list to serialize to the api view
    """
    tasks = []
    url = flask.request.host_url
    for i, task in enumerate(app.queue.list(executor)):
        tasks.append(
            {
                "taskid": task[0],
                "command": task[1],
                "html_url": "%sexecutor/%s" % (url, task[0]),
                "api_url": "%sapi/tasks/%s" % (url, task[0]),
            }
        )
    return tasks


class apiList(Resource):
    """display all tasks
    """

    def get(self):
        return list_tasks()


class apiListExecutor(Resource):
    """display all tasks for an executor
    """

    def get(self, executor):
        return list_tasks(executor)


class apiGet(Resource):
    """display a specific task
    :param taskid: id for a specific task
    """

    def get(self, taskid):
        task = app.queue.get(taskid)
        return task.export()


api = Api(app)
api.add_resource(apiList, "/api/tasks")
api.add_resource(apiGet, "/api/tasks/<string:taskid>")
api.add_resource(apiListExecutor, "/api/tasks/executor/<string:executor>")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""


def confirm(prompt, response=False):
    """Used to prompt the user for a yes or no response, and returns True/False.
       This means we can use it like:
       if confirm("Would you like to do the thing?"):
          ....
    """
    prompt = "%s [%s]|%s: " % (prompt, "n", "y")
    while True:
        answer = input(prompt)
        if not answer:
            return response
        if answer not in ["y", "Y", "n", "N"]:
            print("Please respond with y or n.")
            continue
        if answer == "y" or answer == "Y":
            return True
        if answer == "n" or answer == "N":
            return False



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import sys


class Database:
    """A qme database holds common functions to represent tasks, and results.
       This is literally the base of the database, so it holds common shared
       functions to generate a task id or similar. The child classes (the
       specific database types) should have an init functions that ensures
       connection or creation of folders is possible.
    """

    database = "notimplemented"

    def clear(self):
        """clear (delete) all tasks.
        """
        raise NotImplementedError

    def search(self, query):
        """search is only available to non-filesystem databases
        """
        raise NotImplementedError

    def list_tasks(self, name=None):
        """list tasks associated with an executor, or all tasks.

           Arguments:
           - executor (str) : the executor type. If not provided, list all
        """
        raise NotImplementedError

    def add_task(self, executor):
        """Create a filesystem task based on an executor type. The executor controls
           what data is exported and the uid, the task object just handles saving it.
        """
        raise NotImplementedError

    def update_task(self, executor, updates=None):
        """update a task with a json dictionary.
        """
        raise NotImplementedError

    def get_task(self, taskid=None):
        """Get a task based on a taskid. Exits on error if doesn't exist. If
           a task id is not provided, get the last run task.
        """
        raise NotImplementedError

    def delete_task(self, taskid):
        """delete a task based on a specific task id. All task ids must be
           in the format of <taskid>-<uid> without extra dashes so we can
           reliably split based on the first dash.
        """
        raise NotImplementedError

    def delete_executor(self, name):
        """delete all tasks for an executor, based on executor's name (str).
        """
        raise NotImplementedError

    def iter_executors(self, fullpath=False):
        """list executors based on the subfolders in the base database folder.
        """
        raise NotImplementedError



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.main import Queue
import sys
import os


def main(args, extra):

    # Create a queue object, run the command to match to an executor
    queue = Queue(config_dir=args.config_dir)

    # Pass the queue object to start a server
    try:
        from qme.app.server import start

        start(port=args.port, queue=queue, debug=args.debug)
    except:
        sys.exit(
            "You must 'pip install qme[app]' 'pip install qme[all]' to use the dashboard."
        )



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from flask import Flask, render_template
from flask_socketio import emit

from qme.app.server import app, socketio
from qme.defaults import QME_SOCKET_UPDATE_SECONDS


## Executor Views


@app.route("/executor/<taskid>")
def shell_executor(taskid):

    # TODO: return executor template based on task.executor.name
    task = app.queue.get(taskid)
    return render_template(
        "executors/shell.html", task=task.load(), database=app.queue.database
    )



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.main import Queue
from qme.logger import bot
import sys


def main(args, extra):

    # Create a queue object
    queue = Queue(config_dir=args.config_dir)
    query = " ".join(args.query).strip()
    if not query:
        sys.exit("Please provide a query to search for.")
    results = queue.search(query)
    bot.table(results)



#!/usr/bin/env python
"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import sys
import pytest


def test_executor_slurm(tmp_path):
    """Test the shell executor. A shell executor should have added a returncode,
       output, and error to it's data export.
    """
    # We can't emulate slurm, so just test general executor
    from qme.main.executor import get_named_executor

    executor = get_named_executor("slurm")
    assert executor.name == "slurm"
    actions = executor.get_actions()
    for key in ["status", "output", "error", "cancel"]:
        assert key in actions

    # Task.export includes the executor specific data
    data = executor.export()
    for key in ["cmd", "pwd", "output", "error", "returncode"]:
        assert key in data



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.defaults import QME_DATABASE
from qme.main.config import Config
from qme.main.database import init_db
from qme.main.executor import get_executor
from qme.utils.regex import uuid_regex
from qme.utils.prompt import confirm
from qme.logger import bot

import os
import re
import sys


class Queue:
    """A Queue Me (qme) Queue is the core operator for the qme library. Here
       we initialize a queue, meaning:
       1. taking and setting up a user specified database or the default
       2. setting up a configuration file in $HOME (or at QME_HOME)
       3. then can be used to show a status, parse a job, etc.
    """

    def __init__(self, config_dir=None, database=None):
        """create a queue. We take a config directory (defaults to $HOME/.qme)
           and a database configuration string (defaults to filesystem to be
           created at $HOME/.qme/database
        """
        self.config = Config(config_dir)
        self.config_dir = os.path.dirname(self.config.configfile)
        self.initdb(database)

    def initdb(self, database):
        """setup the qme home (where the config directory is stored) and the
           database specification. If a database string is required (and not
           provided) alert the user and exit on error).

           Arguments:
            - config_dir (str) : the configuration directory (home for qme)
            - database (str) : a string to specify the database setup
        """
        self.database = (
            database
            or QME_DATABASE
            or self.config.get("DEFAULT", "database")
            or "filesystem"
        )
        database_string = self.config.get("DEFAULT", "databaseconnect")
        bot.info("Database: %s" % self.database)

        # Supported database options
        valid = ("sqlite", "postgresql", "mysql+pymysql", "filesystem")
        if not self.database.startswith(valid):
            bot.warning(
                "%s is not yet a supported type, saving to filesystem." % self.database
            )
            self.database = "filesystem"

        # Create database client with functions for database type
        self.db = init_db(
            self.database,
            config_dir=self.config_dir,
            database_string=database_string,
            config=self.config,
        )

    def list(self, executor=None):
        """A wrapper to the database list_tasks function. Optionally take
           a whole executor name (e.g., shell) or just a specific task. No
           executor indicates that we list everything.
        """
        return self.db.list_tasks(executor)

    def get(self, taskid=None):
        """A wrapper to get a task id from the database. If an id is not provided,
           will return the last run task based on timestamp of file or database.
        """
        return self.db.get_task(taskid)

    def clear(self, target=None, noprompt=False):
        """clear takes a target, and that can be a taskid, executor, or none
           We ask the user for confirmation.
        """
        # Case 1: no target indicates clearing all
        if not target:
            if noprompt or confirm(f"This will delete all tasks, are you sure?"):
                return self.db.clear()

        # Case 2, it's a specific taskid
        elif re.search(uuid_regex, target):
            if noprompt or confirm(f"This will delete task {target}, are you sure?"):
                return self.db.delete_task(target)

        # Case 2: it's an executor
        elif target in list(self.db.iter_executors()):
            if noprompt or confirm(
                f"This will delete all executor {target} tasks, are you sure?"
            ):
                return self.db.delete_executor(target)
        else:
            sys.exit(f"Unrecognized target to clear {target}")

    def run(self, command, message=None):
        """Given a command, get the executor for it (also creating an entry
           in the task database) and run the command.
        """
        executor = get_executor(command, config=self.config)

        # add executor unique id and command to the database, returns a task object
        task = self.db.add_task(executor)

        # Execute and store result (this will need to be generalized)
        task.executor.execute(message=message)
        self.db.update_task(task.executor)
        bot.info(f"{task.summary()}")
        return task

    def rerun(self, taskid=None, message=None):
        """Given a command, get the executor for it (also creating an entry
           in the task database) and run the command. If the task is found and
           rerun, it is returned. Otherwise None is returned.
        """
        task = self.db.get_task(taskid)
        params = task.load()
        command = params.get("command")
        pwd = params.get("pwd")
        if pwd:
            os.chdir(pwd)
        if command:
            task.executor.execute(command, message=message)
            self.db.update_task(task.executor)
            bot.info(f"{task.summary()}")
            return task
        else:
            bot.warning(f"{task.executor.taskid} does not have an associated command.")

    def search(self, query):
        """Search across commands and general metadata for a string of interest.
           We use regular expressions (re.search) so they are supported.
           Search is only available for non-filesystem databases.
        """
        if self.database == "filesystem":
            sys.exit("Search is only supported for relational databases.")
        return self.db.search(query)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from qme.utils.file import get_userhome
import multiprocessing
import os
import sys


def getenv(variable_key, default=None, required=False, silent=True):
    """ attempt to get an environment variable. If the variable
        is not found, None is returned.

        Arguments:

         - variable_key (str) : the variable name
         - required (bool) : exit with error if not found
         - silent (bool) : Do not print debugging information
    """
    variable = os.environ.get(variable_key, default)
    if variable is None and required:
        bot.error("Cannot find environment variable %s, exiting." % variable_key)
        sys.exit(1)

    if not silent and variable is not None:
        bot.verbose("%s found as %s" % (variable_key, variable))

    return variable


QME_NPROC = multiprocessing.cpu_count()
QME_WORKERS = int(getenv("QME_WORKERS", QME_NPROC * 2 + 1))
QME_SHELL = getenv("QME_SHELL", "ipython")

# Default database is filesystem
QME_DATABASE = getenv("QME_DATABASE")

# Determine database backend to use, and where to store config
USERHOME = get_userhome()
QME_HOME = os.path.join(USERHOME, ".qme")

# Database folder for filesystem or sqlite database
QME_DATABASE_STRING = os.environ.get("QME_DATABASE")

# Dashboard settings
QME_SOCKET_UPDATE_SECONDS = int(getenv("QME_SOCKET_UPDATE_SECONDS", 5))
QME_HOSTNAME = getenv("QME_HOSTNAME", "127.0.0.1")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""


def init_db(database, config_dir=None, database_string="", config=None):
    """Initialize the database, meaning a base client and appropriate functions
       to save, or generate a unique ID based on the backend being used. Each
       client has it's own init to check for a connection (or filesystem 
       path existence) and then functions to interact with entities.
    """
    # Case 1: Filesystem database saves to $HOME/.qme/database
    if database == "filesystem":
        from .filesystem import FileSystemDatabase as Database

    # Case 2: Sqlite database saves to $HOME/.qme/qme.db
    elif database.startswith("sqlite"):
        from .sqlite import SqliteDatabase as Database

    # Case 3: Postgresql or mysql+pymysql
    else:
        from .relational import RelationalDatabase as Database
    return Database(
        config_dir=config_dir,
        database_string=database_string,
        database=database,
        config=config,
    )



#!/usr/bin/env python

"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import qme
import argparse
import sys
import os


def get_parser():
    parser = argparse.ArgumentParser(description="QueueMe job executor and dashboard.")

    parser.add_argument(
        "--version",
        dest="version",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--config_dir",
        dest="config_dir",
        help="select database and configuration directory (defaults to $HOME/.qme).",
    )

    parser.add_argument(
        "-m",
        "--message",
        dest="message",
        help="Add a message or description to store alongside a run.",
    )

    description = "actions for qme"
    subparsers = parser.add_subparsers(
        help="qme actions", title="actions", description=description, dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", help="show software version")

    # Configure qme (not written yet, will be written when databases added)
    config = subparsers.add_parser("config", help="configure qme.")

    # Specify a database, if not sqlite must include a complete string
    config.add_argument(
        "--db",
        "--database",
        dest="database",
        default=None,
        help="select database backend for qme. [filesystem|sqlite] or [sqlite|mysql|postgresql]:///",
    )

    config.add_argument(
        "--set",
        help="set a setting, provide the executor, key, and value (set slurm sacct_format value)",
        default=None,
        dest="set",
        nargs=3,
    )

    # Clear an entire executor family, one task, or all tasks
    clear = subparsers.add_parser("clear", help="Run a command to add to the queue.")
    clear.add_argument("target", nargs="?")
    clear.add_argument(
        "--force",
        dest="force",
        help="Don't ask for confirmation for delete (for headless).",
        default=False,
        action="store_true",
    )

    execute = subparsers.add_parser(
        "exec", help="Execute an action for the last task, or a taskid"
    )
    execute.add_argument("actions", nargs="*")

    generate = subparsers.add_parser(
        "generate-key",
        help="generate a key for qme start, should be exported to QME_SERVER_KEY.",
    )

    # List tasks and print to terminal
    ls = subparsers.add_parser("ls", help="List tasks")
    ls.add_argument(
        "executor", help="list one or more executors or taskids.", nargs="*"
    )

    # Run a command (gets passed to executor via template)
    run = subparsers.add_parser("run", help="Run a command.")
    run.add_argument(
        "cmd",
        nargs="*",
        help="The command to parse. If --help needed, should be fully quoted.",
    )

    # Rerun a task
    rerun = subparsers.add_parser("rerun", help="Re-run a particular task.")
    rerun.add_argument("taskid", nargs="?")

    # Start the queueMe dashboard
    search = subparsers.add_parser(
        "search",
        help="Search for content in a command or metadata (sqlite or relational only).",
    )
    search.add_argument("query", nargs="*")

    # Start the queueMe dashboard
    start = subparsers.add_parser(
        "start", help="View the queue web interface (requires Flask)"
    )
    start.add_argument(
        "--port",
        dest="port",
        default=5000,
        type=int,
        help="select port to run qme dashboard on (defaults to 5000)",
    )
    start.add_argument(
        "--host",
        dest="host",
        default="127.0.0.1",
        type=str,
        help="the hostname to run for the server (defaults to 127.0.0.1)",
    )

    start.add_argument(
        "--debug",
        dest="debug",
        help="run server in debug mode (defaults to False)",
        default=False,
        action="store_true",
    )

    # Print complete metadata for a specific task
    get = subparsers.add_parser("get", help="Get task")
    get.add_argument("taskid", help="list taskid to return", nargs="?")
    return parser


def main():
    """main entrypoint for qme
    """

    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client 
           and exit with return code.
        """
        version = qme.__version__

        print("\nQueueMe Python v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    # Show the version and exit
    if args.command == "version" or args.version:
        print(qme.__version__)
        sys.exit(0)

    # Does the user want a shell?
    if args.command == "clear":
        from .clear import main
    if args.command == "config":
        from .config import main
    if args.command == "exec":
        from .actions import main
    if args.command == "generate-key":
        from .generate import main
    if args.command == "get":
        from .get import main
    if args.command == "ls":
        from .listing import main
    if args.command == "run":
        from .run import run as main
    if args.command == "rerun":
        from .run import rerun as main
    if args.command == "search":
        from .search import main
    if args.command == "start":
        from .start import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, extra=extra)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    main()
